use axum::{
    body::Body,
    extract::{Path, State},
    response::Redirect,
    routing::get,
    BoxError, Router,
};
use bytes::Bytes;
use http::{Request, Response};
use http_body_util::{combinators::UnsyncBoxBody, BodyExt};
use mime::Mime;
use sqlx::query;
use tower_http::services::ServeFile;
use ulid::Ulid;
use uuid::Uuid;

use super::SharedState;
use crate::error::AppError;

pub fn router() -> Router<SharedState> {
    Router::new()
        .route("/:slug", get(redirect_link))
        .route("/f/:key", get(download_file))
}

async fn redirect_link(
    State(SharedState { db, .. }): State<SharedState>,
    Path(slug): Path<String>,
) -> Result<Redirect, AppError> {
    match query!("SELECT destination FROM link WHERE slug = $1", slug)
        .fetch_optional(&db)
        .await?
    {
        Some(r) => Ok(Redirect::temporary(&r.destination)),
        None => Err(AppError::LinkNotFoundSlug(slug)),
    }
}

async fn download_file(
    State(SharedState { db, config }): State<SharedState>,
    Path(key): Path<Ulid>,
    request: Request<Body>,
) -> Result<Response<UnsyncBoxBody<Bytes, BoxError>>, AppError> {
    match query!(
        "SELECT hash, mime FROM file_key JOIN file ON file_id = file.id WHERE file_key.id = $1",
        Uuid::from(key),
    )
    .fetch_optional(&db)
    .await?
    .map(|r| (r.hash, r.mime))
    {
        Some((hash, mime)) => {
            let mime: Option<Mime> = mime.parse().ok();
            let path = config.file_store_dir.join(hex::encode(hash));
            let mut sf = match mime {
                Some(mime) => ServeFile::new_with_mime(path, &mime),
                None => ServeFile::new(path),
            };
            match sf.try_call(request).await {
                Ok(response) => Ok(response.map(|body| body.map_err(Into::into).boxed_unsync())),
                Err(err) => Err(AppError::Io(err)),
            }
        }
        None => Err(AppError::FileKeyNotFound(key)),
    }
}
